package org.bdware.server.irp;

import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpHandler;
import org.apache.commons.io.IOUtils;
import org.bdware.sc.util.JsonUtil;
import org.rocksdb.RocksDB;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

public class RegisterHandler implements HttpHandler {
    RocksDB db;
    String dbPath = "./handleRecords/";

    public RegisterHandler(RocksDB rocksDB) {
        db = rocksDB;
    }

    public Map<String, String> HandleRequest(Map<String, String> postInfo) {
        Map<String, String> respMap = new HashMap<>();
        String action = postInfo.get("action");
        switch (action) {
            case "resolve":
                respMap = handleResolve(postInfo);
                break;
            case "register":
            case "forceRegister":
                respMap = handleRegister(postInfo);
                break;
            case "unregister":
            case "forceDelete":
                respMap = handleUnRegister(postInfo);
                break;
            case "reregister":
                respMap = handleReregister(postInfo);
                break;
            default:
                System.out.println("error response");
        }
        return respMap;
    }

    public Map<String, String> handleRegister(Map<String, String> reqMap) {
        Map<String, String> respMap = new HashMap<>();
        String type;
        switch (reqMap.get("hrType")) {
            case "do":
                type = "do.";
                ViewHandler.handleRecordCount.doCount++;
                break;
            case "dou":
                type = "dou.";
                ViewHandler.handleRecordCount.userCount++;
                break;
            case "doip":
                type = "doip.";
                ViewHandler.handleRecordCount.repoCount++;
                break;
            default:
                type = "ukw.";
                break;
        }
        String identifier = LocalLHSProxy.PREFIX + type + geneRandomID() + "_bdw";
        reqMap.remove("action");
        reqMap.remove("hrType");
        reqMap.put("identifier", identifier);
        String handleRecord = JsonUtil.toJson(reqMap);
        try {
            db.put(identifier.getBytes(), handleRecord.getBytes());
            System.out.println("successful update do, identifier: " + identifier);
            respMap.put("identifier", identifier);
            respMap.put("status", "1");
            respMap.put("response", "register DO success, identifier: " + identifier);
            ViewHandler.addRegister();
        } catch (Exception e) {
            e.printStackTrace();
            respMap.put("status", "0");
            respMap.put("response", "respond failed from LHS");
        }
        return respMap;
    }

    public Map<String, String> handleReregister(Map<String, String> reqMap) {
        Map<String, String> respMap = new HashMap<>();
        String identifier;
        if (reqMap.get("identifier") != null) identifier = reqMap.get("identifier");
        else {
            respMap.put("status", "0");
            respMap.put("response", "identifier not found");
            return respMap;
        }
        try {
            reqMap.remove("action");
            reqMap.remove("hrType");
            String handleRecord = JsonUtil.toJson(reqMap);
            if (db.get(identifier.getBytes()) == null) {
                String idType;
                if ((identifier.split("/")).length > 2)
                    idType = identifier.split("/")[1].split("\\.")[0];
                else
                    idType = identifier.split("\\.")[0];
                switch (idType) {
                    case "do":
                        ViewHandler.handleRecordCount.doCount++;
                        break;
                    case "dou":
                        ViewHandler.handleRecordCount.userCount++;
                        break;
                    case "doip":
                        ViewHandler.handleRecordCount.repoCount++;
                        break;
                    default:
                        break;
                }
            } else {
                db.delete(identifier.getBytes());
            }

            db.put(identifier.getBytes(), handleRecord.getBytes());
            System.out.println("successful update do, identifier: " + identifier);
            respMap.put("identifier", identifier);
            respMap.put("status", "1");
            respMap.put("response", "reRegister DO success, identifier: " + identifier);
        } catch (Exception e) {
            e.printStackTrace();
            respMap.put("status", "0");
            respMap.put("response", "unregister failed: " + e.getMessage());
        }
        return respMap;
    }

    public Map<String, String> handleUnRegister(Map<String, String> reqMap) {
        Map<String, String> respMap = new HashMap<>();
        String identifier;
        if (reqMap.get("identifier") != null) identifier = reqMap.get("identifier");
        else {
            respMap.put("status", "0");
            respMap.put("response", "identifier not found");
            return respMap;
        }
        try {
            switch (identifier.split("\\.")[2]) {
                case "470/do":
                    ViewHandler.handleRecordCount.doCount--;
                    break;
                case "470/doip":
                    ViewHandler.handleRecordCount.repoCount--;
                    break;
                case "470/dou":
                    ViewHandler.handleRecordCount.userCount--;
                    break;
            }
            db.delete(identifier.getBytes());
            respMap.put("status", "1");
            respMap.put("response", "success delete: " + identifier);
            ViewHandler.addDelete();
        } catch (Exception e) {
            e.printStackTrace();
            respMap.put("status", "0");
            respMap.put("response", "unregister failed: " + e.getMessage());
        }
        return respMap;
    }

    public Map<String, String> handleResolve(Map<String, String> reqMap) {
        Map<String, String> respMap = new HashMap<>();
        String identifier;
        if (reqMap.get("identifier") != null) identifier = reqMap.get("identifier");
        else {
            respMap.put("status", "0");
            respMap.put("response", "identifier not found");
            return respMap;
        }
        try {
            if (db.get(identifier.getBytes()) == null) {
                respMap.put("status", "0");
                respMap.put("response", "not exist");
                return respMap;
            }
            String result = new String(db.get(identifier.getBytes()));
            respMap = JsonUtil.fromJson(result, HashMap.class);
            ViewHandler.addResolve();
        } catch (Exception e) {
            e.printStackTrace();
            respMap.put("status", "0");
            respMap.put("response", "resolve failed: " + e.getMessage());
        }
        return respMap;
    }

    public String geneRandomID() {
        String str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
        Random random = new Random();
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < 10; i++) {
            int number = random.nextInt(62);
            sb.append(str.charAt(number));
        }
        return sb.toString();
    }

    @Override
    public void handle(HttpExchange exchange) throws IOException {
        Map<String, String> respMap;
        try {
            //获得表单提交数据(post)
            String postString = IOUtils.toString(exchange.getRequestBody());
            Map<String, String> postInfo = formData2Dic(postString);
            respMap = HandleRequest(postInfo);
            exchange.sendResponseHeaders(200, 0);
            OutputStream os = exchange.getResponseBody();
            os.write(JsonUtil.toJson(respMap).getBytes());
            os.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public Map<String, String> formData2Dic(String formData) {
        Map<String, String> result = new HashMap<>();
        if (formData == null || formData.trim().length() == 0) {
            return result;
        }
        final String[] items = formData.split("&");
        Arrays.stream(items).forEach(item -> {
            final String[] keyAndVal = item.split("=");
            if (keyAndVal.length == 2) {
                try {
                    final String key = URLDecoder.decode(keyAndVal[0], "utf8");
                    final String val = URLDecoder.decode(keyAndVal[1], "utf8");
                    result.put(key, val);
                } catch (UnsupportedEncodingException e) {
                    e.printStackTrace();
                }
            }
        });
        return result;
    }
}
