package org.bdware.server.nodecenter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.bean.ContractDesp;
import org.bdware.sc.db.KeyValueDBUtil;

import java.util.List;

public class CMNode {
    private static final Logger LOGGER = LogManager.getLogger(CMNode.class);
    // public List<ContractDesp> contracts;
    public volatile List<ContractDesp> contracts; // 有路由信息
    public transient NodeCenterActions connection;
    public String pubKey;
    public String nodeName;
    public String udpID;
    public String ipPort;
    public int events;
    public String cimanager = "";
    public String peerID;
    public String masterAddress; // 这个节点如果作为master的地址，例如18010的这个值就是18011
    public int contractVersion;

    public CMNode(NodeCenterActions nodeCenterController, String pubKey) {
        connection = nodeCenterController;
        this.pubKey = pubKey;
    }

    // TODO nodeMangerPubkey
    public void updateContract(List<ContractDesp> cons, int version) {
        contracts = cons;
        contractVersion = version;
        for (ContractDesp desp : cons) {
            KeyValueDBUtil.instance.setValue(
                    NCTables.ContractMeta.toString(), desp.contractID, desp.contractName);
        }
    }

    public String formatContractName(String contractIDOrName) {
        if (null != contracts) {
            for (ContractDesp desp : contracts) {
                if (desp.contractID.equals(contractIDOrName)
                        || desp.contractName.equals(contractIDOrName)) {
                    return desp.contractName;
                }
            }
        }
        return null;
    }

    public boolean containsContract(String contractIDOrName) {
        if (null != contracts) {
            for (ContractDesp desp : contracts) {
                if (desp.contractID.equals(contractIDOrName)
                        || desp.contractName.equals(contractIDOrName)) return true;
            }
        }
        return false;
    }

    public boolean containsEvent(String contractIDOrName, String event) {
        if (null != contracts) {
            for (ContractDesp desp : contracts) {
                if (desp.contractID.equals(contractIDOrName)
                        || desp.contractName.equals(contractIDOrName)) {
                    if (desp.events.containsKey(event)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public boolean checkAlive() {
        if (null == connection) {
            return false;
        }
        try {
            if (!connection.controller.isOpen()) {
                LOGGER.info("node " + nodeName + "(" + pubKey.substring(0, 5) + ") may be offline!");
//                System.out.println(
//                        new SimpleDateFormat("yyyy-MM-dd HH:mm:ss")
//                                .format(new Date(System.currentTimeMillis()))
//                                + "[ADSP][CMNode] isAlive : "
//                                + nodeName
//                                + "  not open!"
//                                + "\n");
                return doubleCheckAlive();
            } else {
                return true;
            }
        } catch (Exception e) {
            LOGGER.error("checking alive failed! " + e.getMessage());
        }
        return false;
    }

    private boolean doubleCheckAlive() {
        try {
            // Thread.sleep(3000L);
            Thread.sleep(500L);
            if (connection.controller.isOpen()) {
                LOGGER.info("node " + nodeName + "(" + pubKey.substring(0, 5) + ") is online!");
                return true;
            }
        } catch (Exception e) {
            LOGGER.error("checking alive failed! " + e.getMessage());
        }
        LOGGER.info("node " + nodeName + "(" + pubKey.substring(0, 5) + ") may be offline!");
        return false;
    }

    public void removeCIManager(String string) {
        LOGGER.debug("removeCIManager" + string);
        int start = this.cimanager.indexOf(string);
        if (start > 0) {
            this.cimanager =
                    this.cimanager
                            .substring(0, start)
                            .concat(this.cimanager.substring(start + 130));
        }
    }

    public void addCIManager(String string) {
        LOGGER.debug("addCIManager" + string);
        this.cimanager = this.cimanager.concat(" " + string);
    }

    public void setCIManager(String string) {
        this.cimanager = string;
    }

    public void setPeerID(String string) {
        this.peerID = string;
    }

    public void setIpPort(String string) {
        this.ipPort = string;
    }
}
