package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.CharArraySet;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.WordlistLoader;
import org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.analysis.tokenattributes.CharTermAttribute;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StringField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.*;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.*;
import org.apache.lucene.store.FSDirectory;
import org.bdware.sc.bean.ContractDesp;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.action.Action;
import org.bdware.server.nodecenter.searchresult.ContractMeta;
import org.bdware.server.nodecenter.searchresult.ResultModel;
import org.wltea.analyzer.lucene.IKAnalyzer;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

public class MetaIndexAction { // public static IndexWriter indexWriter;
    private static final Logger LOGGER = LogManager.getLogger(MetaIndexAction.class);
    // public static IndexSearcher indexSearcher;
    // public static TopDocs topDocs;
    // public static Query query;
    // public static QueryParser queryParser;
    // public static Analyzer analyzer;
    // public static IndexWriterConfig config;
    // public static List<Document> docList = new ArrayList<>();
    public static FSDirectory indexDir;
    public static Integer PAGE_SIZE = 10;
    public static Integer page;
    public static boolean isEmpty = false;
    public static NodeCenterFrameHandler controller;
    //    public MetaIndexAction(NodeCenterFrameHandler nodeCenterFrameHandler) {
    //        controller = nodeCenterFrameHandler;
    //    }
    private static IndexWriter indexWriter;

    static {
        initIndex();
    }

    private static void initIndex() {
        try {
            File dir = new File("./NodeCenterDB/MetaIndex");
            if (!dir.exists()) {
                LOGGER.info("make metaIndex dir ");
                dir.mkdirs();
                isEmpty = true;
            }
            indexDir = FSDirectory.open(Paths.get(dir.toURI()));
            // Analyzer analyzer = new StandardAnalyzer();
            Analyzer analyzer = new IKAnalyzer();
            IndexWriterConfig config = new IndexWriterConfig(analyzer);
            indexWriter = new IndexWriter(indexDir, config);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void updateContractsIndex(List<ContractDesp> contracts, ResultCallback rc)
            throws IOException {
        IndexReader indexReader = null;
        for (ContractDesp thisDesp : contracts) {
            JsonObject req = new JsonObject();
            if (!isEmpty) {
                try {
                    indexReader = DirectoryReader.open(indexDir);
                    IndexSearcher indexSearcher = new IndexSearcher(indexReader);
                    LOGGER.debug(thisDesp.contractName + "--> try to index");
                    Query query = new TermQuery(new Term("contractID", thisDesp.contractID));
                    TopDocs docs = indexSearcher.search(query, 10);
                    LOGGER.debug(docs.scoreDocs);
                    if (null != thisDesp.contractName && (docs.scoreDocs == null || docs.scoreDocs.length == 0)) {
                        req.addProperty("action", "requestReadMe");
                        req.addProperty("contractID", thisDesp.contractID);
                        rc.onResult(req);
                        LOGGER.info("contract " + thisDesp.contractName + " --> actually to index");
                        continue;
                    }
                } catch (Exception e) {
                    LOGGER.warn("getting index failed! " + e.getMessage());
                }
            }
            req.addProperty("action", "requestReadMe");
            req.addProperty("contractID", thisDesp.contractID);
            rc.onResult(req);
            LOGGER.info("contract " + thisDesp.contractName + " --> actually to index");
        }
        if (null != indexReader) {
            indexReader.close();
        }
    }

    public static String accuSearch(String keyword) {
        try {
            DirectoryReader indexReader = DirectoryReader.open(indexDir);
            IndexSearcher indexSearcher = new IndexSearcher(indexReader);
            Term t = new Term("contractID", keyword);
            Query query = new TermQuery(t);
            TopDocs docs = indexSearcher.search(query, 10);
            return docs.scoreDocs.length + "";
        } catch (Exception e) {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            return bo.toString();
        }
    }

    public static void delIndexbyCID(JsonObject jo) {
        try {
            if (!jo.has("contractID")) return;
            String contractID = jo.get("contractID").getAsString();
            LOGGER.info("contractID:" + contractID + "-->actually to delete");
            indexWriter.deleteDocuments(new Term("contractID", contractID));
            indexWriter.commit();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static void delIndexbyOwner(JsonObject jo) {
        try {
            if (!jo.has("owner")) return;
            String owner = jo.get("owner").getAsString();
            LOGGER.info("owner:" + owner + "-->actually to delete");
            indexWriter.deleteDocuments(new Term("owner", owner));
            indexWriter.commit();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static String search(JsonObject searchReq) {
        try {
            if (!searchReq.has("searchType")) {
                return "missing arguments";
            }
            switch (searchReq.get("searchType").getAsString()) {
                case "getMetabyCID":
                    return getMetabyCID(searchReq);
                case "getMetabyOwner":
                    return getMetabyOwner(searchReq);
                case "getMetabyPubkey":
                    return getMetabyPubkey(searchReq);
                case "getMetabyReadme":
                    return getMetabyReadme(searchReq);
                default:
                    return "no such search type";
            }
            //            if (searchReq != null) {
            //                logger.info(searchReq.get("pubkey").getAsString() + "---->start
            // search");
            //                DirectoryReader indexReader = DirectoryReader.open(indexDir);
            //                IndexSearcher indexSearcher = new IndexSearcher(indexReader);
            //                BooleanQuery.Builder queryBuilder = new BooleanQuery.Builder();
            //                reandler(searchReq, queryBuilder);
            //                BooleanQuery searchQuery = queryBuilder.build();
            //                TopDocs docs = indexSearcher.search(searchQuery, 10);
            //                return docs.scoreDocs.length + "";
            //            }
            //            else return "0";
        } catch (Exception e) {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            return bo.toString();
        }
    }

    public static String getMetabyReadme(JsonObject jo) {
        try {
            if (!jo.has("keyword")) return "missing arguments: keyword";

            if (!jo.has("page")) page = 1;
            else page = jo.get("page").getAsInt();
            if (page <= 0) page = 1;
            if (jo.has("pageSize")) PAGE_SIZE = jo.get("pageSize").getAsInt();
            String keyword = jo.get("keyword").getAsString();
            // Analyzer analyzer = new StandardAnalyzer();
            Analyzer analyzer = new IKAnalyzer();
            QueryParser queryParser = new QueryParser("readmeStr", analyzer);
            Query rmQuery = queryParser.parse(keyword);

            DirectoryReader indexReader = DirectoryReader.open(indexDir);
            IndexSearcher indexSearcher = new IndexSearcher(indexReader);
            TopDocs docs = indexSearcher.search(rmQuery, 1000);
            //            Document document = null ;
            //            if (docs.scoreDocs!=null&& docs.scoreDocs.length>0) {
            //                document = indexSearcher.doc(docs.scoreDocs[0].doc);
            //                ContractMeta contractMeta = new ContractMeta();
            //                contractMeta.setContractID(document.get("contractID"));
            //                contractMeta.setOwner(document.get("owner"));
            //                contractMeta.setPubkey(document.get("pubkey"));
            //                contractMeta.setReadmeStr(document.get("readmeStr"));
            //                String rs = JsonUtil.toJson(contractMeta);
            //                return rs;
            //            }
            ResultModel resultModel = null;
            if (docs.scoreDocs != null && docs.scoreDocs.length > 0)
                resultModel = paginate(docs, indexReader);
            if (resultModel != null) {
                String rs = JsonUtil.toJson(resultModel);
                return rs;
            }
            return "not found";
        } catch (Exception e) {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            return bo.toString();
        }
    }

    private static String getMetabyPubkey(JsonObject jo) {
        try {
            if (!jo.has("pubkey")) return "missing arguments: pubkey";

            if (!jo.has("page")) page = 1;
            else page = jo.get("page").getAsInt();
            if (page <= 0) page = 1;

            String pubkey = jo.get("pubkey").getAsString();
            DirectoryReader indexReader = DirectoryReader.open(indexDir);
            IndexSearcher indexSearcher = new IndexSearcher(indexReader);
            Query query = new TermQuery(new Term("pubkey", pubkey));
            TopDocs docs = indexSearcher.search(query, 10);
            //            Document document = null;
            //            if (docs.scoreDocs!=null&& docs.scoreDocs.length>0) {
            //                document = indexSearcher.doc(docs.scoreDocs[0].doc);
            //                ContractMeta contractMeta = new ContractMeta();
            //                contractMeta.setContractID(document.get("contractID"));
            //                contractMeta.setOwner(document.get("owner"));
            //                contractMeta.setPubkey(document.get("pubkey"));
            //                contractMeta.setReadmeStr(document.get("readmeStr"));
            //                String rs = JsonUtil.toJson(contractMeta);
            //                return rs;
            //            }
            ResultModel resultModel = null;
            if (docs.scoreDocs != null && docs.scoreDocs.length > 0)
                resultModel = paginate(docs, indexReader);
            if (resultModel != null) {
                String rs = JsonUtil.toJson(resultModel);
                return rs;
            }
            return "not found";
        } catch (Exception e) {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            return bo.toString();
        }
    }

    private static String getMetabyCID(JsonObject jo) {
        System.out.println("getMetabyCID" + jo.toString());
        try {
            if (!jo.has("contractID")) return "missing arguments: contractID";

            if (!jo.has("page")) page = 1;
            else page = jo.get("page").getAsInt();
            if (page <= 0) page = 1;

            String contractID = jo.get("contractID").getAsString();
            DirectoryReader indexReader = DirectoryReader.open(indexDir);
            IndexSearcher indexSearcher = new IndexSearcher(indexReader);
            Query query = new TermQuery(new Term("contractID", contractID));
            TopDocs docs = indexSearcher.search(query, 10);
            //            Document document = null;
            //            if (docs.scoreDocs!=null&& docs.scoreDocs.length>0) {
            //                document = indexSearcher.doc(docs.scoreDocs[0].doc);
            //                ContractMeta contractMeta = new ContractMeta();
            //                contractMeta.setContractID(document.get("contractID"));
            //                contractMeta.setOwner(document.get("owner"));
            //                contractMeta.setPubkey(document.get("pubkey"));
            //                contractMeta.setReadmeStr(document.get("readmeStr"));
            //                String rs = JsonUtil.toJson(contractMeta);
            //                return rs;
            //            }
            ResultModel resultModel = null;
            if (docs.scoreDocs != null && docs.scoreDocs.length > 0)
                resultModel = paginate(docs, indexReader);
            if (resultModel != null) {
                String rs = JsonUtil.toJson(resultModel);
                return rs;
            }
            return "not found";
        } catch (Exception e) {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            return bo.toString();
        }
    }

    private static String getMetabyOwner(JsonObject jo) {
        try {
            if (!jo.has("owner")) return "missing arguments: owner";

            if (!jo.has("page")) page = 1;
            else page = jo.get("page").getAsInt();
            if (page <= 0) page = 1;
            if (jo.has("pageSize")) PAGE_SIZE = jo.get("pageSize").getAsInt();
            String owner = jo.get("owner").getAsString();
            DirectoryReader indexReader = DirectoryReader.open(indexDir);
            IndexSearcher indexSearcher = new IndexSearcher(indexReader);
            Query query = new TermQuery(new Term("owner", owner));
            TopDocs docs = indexSearcher.search(query, 1000);
            //            Document document = null;
            //            if (docs.scoreDocs!=null&& docs.scoreDocs.length>0) {
            //                document = indexSearcher.doc(docs.scoreDocs[0].doc);
            //                ContractMeta contractMeta = new ContractMeta();
            //                contractMeta.setContractID(document.get("contractID"));
            //                contractMeta.setOwner(document.get("owner"));
            //                contractMeta.setPubkey(document.get("pubkey"));
            //                contractMeta.setReadmeStr(document.get("readmeStr"));
            //                String rs = JsonUtil.toJson(contractMeta);
            //                return rs;
            //            }
            ResultModel resultModel = null;
            if (docs.scoreDocs != null && docs.scoreDocs.length > 0)
                resultModel = paginate(docs, indexReader);
            if (resultModel != null) {
                String rs = JsonUtil.toJson(resultModel);
                return rs;
            }
            return "not found";
        } catch (Exception e) {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            return bo.toString();
        }
    }

    public static void reqHandler(JsonObject req, BooleanQuery.Builder query)
            throws ParseException {
        if (req.get("contractID").getAsString() != null) {
            query.add(
                    new FuzzyQuery(new Term("contractID", req.get("contractID").getAsString())),
                    BooleanClause.Occur.SHOULD);
        }
        if (req.get("owner").getAsString() != null) {
            query.add(
                    new FuzzyQuery(new Term("owner", req.get("owner").getAsString())),
                    BooleanClause.Occur.SHOULD);
        }
        if (req.get("pubkey").getAsString() != null) {
            query.add(
                    new PrefixQuery(new Term("pubkey", req.get("pubkey").getAsString())),
                    BooleanClause.Occur.SHOULD);
        }
        if (req.get("readmeStr").getAsString() != null) {
            Analyzer analyzer = new StandardAnalyzer();
            QueryParser queryParser = new QueryParser("name", analyzer);
            Query rmQuery = queryParser.parse(req.get("readmeStr").getAsString());
            query.add(rmQuery, BooleanClause.Occur.SHOULD);
        }
    }

    public static ResultModel paginate(TopDocs docs, IndexReader reader) throws IOException {
        ResultModel resultModel = new ResultModel();
        Integer start = (page - 1) * PAGE_SIZE;
        Integer end = page * PAGE_SIZE;
        resultModel.setContractCount(docs.totalHits.value);
        ScoreDoc[] scoreDocs = docs.scoreDocs;
        List<ContractMeta> contractMetaList = new ArrayList<>();
        if (scoreDocs != null) {
            for (int i = start; i < scoreDocs.length && i < end; i++) {
                Document document = reader.document(scoreDocs[i].doc);
                ContractMeta contractMeta = new ContractMeta();
                contractMeta.setContractID(document.get("contractID"));
                contractMeta.setStatus(document.get("status"));
                contractMeta.setOwner(document.get("owner"));
                contractMeta.setPubkey(document.get("pubkey"));
                contractMeta.setReadmeStr(document.get("readmeStr"));
                contractMeta.setNodeAddr(document.get("nodeAddr"));
                contractMeta.setPngUrl(document.get("pngUrl"));
                contractMeta.setName(document.get("name"));
                contractMeta.setDoi(document.get("doi"));
                contractMeta.setBuildTime(Long.parseLong(document.get("buildTime")));
                contractMeta.setDoip("DOIP://86.470.5000/" + document.get("name"));
                contractMetaList.add(contractMeta);
            }
        }
        resultModel.setContractMetaList(contractMetaList);
        resultModel.setCurPage(page);
        Long pageCount =
                docs.totalHits.value % PAGE_SIZE > 0
                        ? (docs.totalHits.value) / PAGE_SIZE + 1
                        : (docs.totalHits.value) / PAGE_SIZE;
        resultModel.setPageCount(pageCount);
        return resultModel;
    }

    @Action(async = true, userPermission = 1L)
    public synchronized void onRequestReadMe(JsonObject json, ResultCallback rc)
            throws IOException {
        //        try {
        //            Thread.sleep(2000);
        //        } catch (InterruptedException e) {
        //            e.printStackTrace();
        //        }
        CMNode node = NodeCenterActions.nodeInfos.get(controller.pubKey);
        String nodeAddr = getNodeAddr(node.masterAddress);
        String name = json.get("name").getAsString();
        // /DOIP/Hello1/assets/logo.png
        // String nodeAddr="127.0.0.1:21030";
        String pngUrl = "http://" + nodeAddr + "/DOIP/" + name + "/assets/logo.png";
        System.out.println(
                "node.masterAddress: "
                        + node.masterAddress
                        + "=========="
                        + "nodeAddr: "
                        + nodeAddr);
        System.out.println("name: " + name + "pngUrl: " + pngUrl);
        String contractID = json.get("contractID").getAsString();
        String status = json.get("status").getAsString();
        String owner = json.get("owner").getAsString();
        String pubkey = json.get("pubkey").getAsString();
        String readmeStr = json.get("readmeStr").getAsString();
        String doi = json.get("doi").getAsString();
        String buildTime = String.valueOf(json.get("buildTime").getAsLong());
        Document document = new Document();
        document.add(new StringField("contractID", contractID, Field.Store.YES));
        document.add(new StringField("status", status, Field.Store.YES));
        document.add(new StringField("owner", owner, Field.Store.YES));
        document.add(new StringField("pubkey", pubkey, Field.Store.YES));
        document.add(new TextField("readmeStr", readmeStr, Field.Store.YES));
        document.add(new StringField("nodeAddr", nodeAddr, Field.Store.YES));
        document.add(new StringField("pngUrl", pngUrl, Field.Store.YES));
        document.add(new StringField("name", name, Field.Store.YES));
        document.add(new StringField("doi", doi, Field.Store.YES));
        document.add(new StringField("buildTime", buildTime, Field.Store.YES));
        LOGGER.info("Index Meta:" + contractID + " str:" + readmeStr);
        LOGGER.info("name:" + name + "=>" + "doi:" + doi + "=>" + "buildTime:" + buildTime);
        indexWriter.addDocument(document);
        indexWriter.commit();
        isEmpty = false;
        // TODO @fanbo 更新lucene
    }

    private String getNodeAddr(String masterAddress) {
        String[] temp = masterAddress.split(":");
        temp[1] = String.valueOf(Integer.parseInt(temp[1]) - 1);
        String nodeAddr = temp[0] + ":" + temp[1];
        return nodeAddr;
    }

    @Action(async = true, userPermission = 1L)
    public void getMetabyReadme(JsonObject json, ResultCallback rc) {
        String result = getMetabyReadme(json);
        JsonObject object = new JsonObject();
        if (json.has("requestID")) object.add("responseID", json.get("requestID"));
        System.out.println("zzzResult" + result);
        object.add("result", JsonParser.parseString(result));
        object.addProperty("action", "getMetabyReadme");
        rc.onResult(object);
    }

    @Action(async = true, userPermission = 1L)
    public void segmentWord(JsonObject json, ResultCallback rc) throws IOException {
        //Analyzer analyzer = new IKAnalyzer(true);
        CharArraySet stopWords = CharArraySet.unmodifiableSet(WordlistLoader.getWordSet(new InputStreamReader(Objects.requireNonNull(MetaIndexAction.class.getClassLoader().getResourceAsStream(
                "org/bdware/server/stopwords.txt")), StandardCharsets.UTF_8)));
        Analyzer analyzer = new SmartChineseAnalyzer(stopWords);
        JsonObject object = new JsonObject();

        String words = "我喜欢区块链，我想试用一下这个，我是做大数据处理的，我是科技局的管理员"; // 从json拿
        TokenStream stream = null;
        if (json.has("requestID")) object.add("responseID", json.get("requestID"));
        try {
            stream = analyzer.tokenStream("myfield", words);
            CharTermAttribute charTermAtt = stream.addAttribute(CharTermAttribute.class);
            stream.reset();
            int count = 0;
            while (stream.incrementToken()) {
                object.addProperty("" + count++, charTermAtt.toString());
                System.out.println(charTermAtt.toString());
            }
            stream.end();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            rc.onResult(object);
            try {
                stream.close();
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
    }

    @Action(async = true, userPermission = 1L)
    public void getMetabyCID(JsonObject json, ResultCallback rc) {
        String result = getMetabyCID(json);
        JsonObject object = new JsonObject();
        if (json.has("requestID")) object.add("responseID", json.get("requestID"));
        object.add("result", JsonParser.parseString(result));
        object.addProperty("action", "getMetabyCID");
        rc.onResult(object);
    }

    @Action(async = true, userPermission = 1L)
    public void getMetabyOwner(JsonObject json, ResultCallback rc) {
        String result = getMetabyOwner(json);
        JsonObject object = new JsonObject();
        if (json.has("requestID")) object.add("responseID", json.get("requestID"));
        object.add("result", JsonParser.parseString(result));
        object.addProperty("action", "getMetabyOwner");
        rc.onResult(object);
    }

    @Action(async = true, userPermission = 1L)
    public void getMetabyPubkey(JsonObject json, ResultCallback rc) {
        String result = getMetabyPubkey(json);
        JsonObject object = new JsonObject();
        if (json.has("requestID")) object.add("responseID", json.get("requestID"));
        object.add("result", JsonParser.parseString(result));
        object.addProperty("action", "getMetabyPubkey");
        rc.onResult(object);
    }

    @Action(async = true, userPermission = 1L)
    public void delIndexbyCID(JsonObject json, ResultCallback rc) {
        delIndexbyCID(json);
    }

    @Action(async = true, userPermission = 1L)
    public void delIndexbyOwner(JsonObject json, ResultCallback rc) {
        delIndexbyOwner(json);
    }
}
