package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufInputStream;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.action.ActionExecutor;

import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class NCClientHandler extends SimpleChannelInboundHandler<Object> {
    private static final Logger LOGGER = LogManager.getLogger(NCClientHandler.class);
    static ExecutorService executorService = Executors.newFixedThreadPool(5);
    public boolean hasPermission;  //是否在目标NC上有权限
    public NCClientActions actions;
    Channel channel;
    ActionExecutor<ResultCallback, JsonObject> ae;
    ChannelHandlerContext ctx;
    private boolean isConnected;

    public NCClientHandler() {
        actions = new NCClientActions(this);
        ae = new ActionExecutor<>(executorService, actions);
        isConnected = false;
        hasPermission = false;
    }

    @Override
    public void channelActive(ChannelHandlerContext ctx) {
        channel = ctx.channel();
        isConnected = true;
        this.ctx = ctx;
        Map<String, String> getSession = new HashMap<>();
        getSession.put("action", "getSessionID");
        sendMsg(JsonUtil.toJson(getSession));
        LOGGER.debug("[NCClientHandler] getSessionID:" + JsonUtil.toJson(getSession));
    }

    @Override
    public void channelInactive(ChannelHandlerContext ctx) {
        isConnected = false;
        hasPermission = false;
        try {
            channel.close();
            channel = null;
        } catch (Exception e) {
        }
    }

    public synchronized void sendMsg(String msg) {
        channel.writeAndFlush(Unpooled.wrappedBuffer(msg.getBytes()));
    }

    @Override
    protected void channelRead0(ChannelHandlerContext ctx, Object msg) throws Exception {
        ByteBuf bb = (ByteBuf) msg;
        try {
            final JsonObject arg =
                    new JsonParser()
                            .parse(new InputStreamReader(new ByteBufInputStream(bb)))
                            .getAsJsonObject();
            if (arg.has("action")) {
                final String action = arg.get("action").getAsString();
                ae.handle(
                        action,
                        arg,
                        new ResultCallback() {
                            @Override
                            public void onResult(String str) {
                                sendMsg(str);
                            }
                        });
            }

        } catch (java.lang.IllegalArgumentException e) {
            LOGGER.debug("[NCClientHandler] can't handle action:" + e.getMessage());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) {
        System.err.println("[NCClientHandler] catchException");
        cause.printStackTrace();
    }

    public void close() {
        try {
            isConnected = false;
            if (channel != null) channel.close();
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            channel = null;
        }
    }

    public boolean isOpen() {
        return ctx != null && ctx.channel().isOpen();
    }
}
